/*
    lm78.c - Part of a Linux module for reading sensor data.
    Copyright (c) 1997, 1998  Alexander Larsson <alla@lysator.liu.se>,
    Frodo Looijaard <frodol@dds.nl> and Philip Edelbrock <phil@netroedge.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/config.h>

/* The next define disables the redefinition of kernel_version. As it is
   already defined and initialized in lm78.c, this would result in twice
   defined symbols. We must include <linux/version.h> by hand now. */
#define __NO_VERSION__

/* Hack to allow the use of CONFIG_MODVERSIONS. In the kernel itself, this
   is arranged from the main Makefile. */
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/module.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/ioport.h>
#include "lm78.h"
#include "main.h"

/* These are functions used only internally. */
static u8 lm78_read_value(u8 address);

static int lm_type = 0;
#define LMTYPE_78 1
#define LMTYPE_78_J 2
#define LMTYPE_79 3

/* This function initializes the LM78. If it is found, and initializing
   succeeds, 0 is returned; if not, an error value is returned. */
int Init_LM78(void)
{
  u8 res;

  if (check_region(LM78_BASE, LM78_EXTENT))
    return -ENODEV;

  outb_p( LM78_DUMMY, LM78_ADDRESS_REG );
  res = inb_p(LM78_ADDRESS_REG);
  inb_p(LM78_DATA_REG); /* reset busy flag */

  if ((res & 0x7F) != LM78_DUMMY)
    return -EIO;

  request_region(LM78_BASE, LM78_EXTENT, "lm78");

  /* Reset all except Watchdog values and last conversion values
     This sets fan-divs to 2, among others */
  LM78_WRITE(LM78_CONFIG, 0x80 );

  res = lm78_read_value(LM78_RESET);
  if (res & 0x80)
     lm_type = LMTYPE_79;
  else if (res & 0x40)
     lm_type = LMTYPE_78_J;
  else
     lm_type = LMTYPE_78;

  /* Start monitoring */
  LM78_WRITE(LM78_CONFIG, (LM78_READ(LM78_CONFIG) & 0xf7) | 0x01);

  if (lm_type == LMTYPE_78) 
    printk("LM78");
  else if (lm_type == LMTYPE_78_J)
    printk("LM78-J");
  else if (lm_type == LMTYPE_79)
    printk("LM79");
  printk(" detected and installed.\n");
#ifdef DEBUG
  printk("LM78 Serial bus address 0x%X.\n",lm78_read_value(LM78_SMBUSADDR));
#endif
  return 0;
}

/* This function cleans up after the module. Only call it after a succesful
   Init_LM78 call! */
void Cleanup_LM78(void)
{
   release_region(LM78_BASE, LM78_EXTENT);
}

/* This function tries to update the current sensor data and to put it
   in human-readable form in buf. It returns the number of characters
   written to buf. 
   TODO: Check length of buf. */
int print_lm78_values (char *buf)
{
  int len = 0;
  int fan, v, max, min;

  /* Redundant test for sensor_data.valid - better safe then sorry! */
  if (sensor_data.valid) {
#define PRINT_SIGN(val) (((val) < 0)?'-':'+')
#define PRINT_VALUES(label, value, max, min) \
    sprintf(buf + len, label ":\t%c%d.%02dV (max = %c%d.%02dV, " \
            "min = %c%d.%02dV)\n", PRINT_SIGN(value), abs(value) / 100, abs(value) % 100, PRINT_SIGN(max), abs(max) / 100,\
            abs(max) % 100, PRINT_SIGN(min),abs(min) / 100, abs(min) % 100)

#ifdef LM78_IN0_STRING
    v = (int)(sensor_data.in[0] * 16 * NORMAL_IN0 / 100000);
    max = (int)(sensor_data.in_max[0] * 16 * NORMAL_IN0 / 100000);
    min = (int)(sensor_data.in_min[0] * 16 * NORMAL_IN0 / 100000);
    len += PRINT_VALUES(LM78_IN0_STRING, v, max, min);
#else
    len += sprintf(buf+len, LM78_IN0_STRING ":\tnot supported\n");
#endif

#ifdef LM78_IN1_STRING
    v = (int)(sensor_data.in[1] * 16 * NORMAL_IN1 / 100000);
    max = (int)(sensor_data.in_max[1] * 16 * NORMAL_IN1 / 100000);
    min = (int)(sensor_data.in_min[1] * 16 * NORMAL_IN1 / 100000);
    len += PRINT_VALUES(LM78_IN1_STRING, v, max, min);
#else
    len += sprintf(buf+len, LM78_IN1_STRING ":\tnot supported\n");
#endif

#ifdef LM78_IN2_STRING
    v = (int)(sensor_data.in[2] * 16 * NORMAL_IN2 / 100000);
    max = (int)(sensor_data.in_max[2] * 16 * NORMAL_IN2 / 100000);
    min = (int)(sensor_data.in_min[2] * 16 * NORMAL_IN2 / 100000);
    len += PRINT_VALUES(LM78_IN2_STRING, v, max, min);
#else
    len += sprintf(buf+len, LM78_IN2_STRING ":\tnot supported\n");
#endif

#ifdef LM78_IN3_STRING
    v = (int)(sensor_data.in[3] * 16 * NORMAL_IN3 / 100000);
    max = (int)(sensor_data.in_max[3] * 16 * NORMAL_IN3 / 100000);
    min = (int)(sensor_data.in_min[3] * 16 * NORMAL_IN3 / 100000);
    len += PRINT_VALUES(LM78_IN3_STRING, v, max, min);
#else
    len += sprintf(buf+len, LM78_IN3_STRING ":\tnot supported\n" );
#endif

#ifdef LM78_IN4_STRING
    v = (int)(sensor_data.in[4] * 16 * NORMAL_IN4 / 100000);
    max = (int)(sensor_data.in_max[4] * 16 * NORMAL_IN4 / 100000);
    min = (int)(sensor_data.in_min[4] * 16 * NORMAL_IN4 / 100000);
    len += PRINT_VALUES(LM78_IN4_STRING, v, max, min);
#else
    len += sprintf(buf+len, LM78_IN4_STRING ":\tnot supported\n" );
#endif

#ifdef LM78_IN5_STRING
    v = (int)(sensor_data.in[5] * 16 * NORMAL_IN5 / 100000);
    max = (int)(sensor_data.in_max[5] * 16 * NORMAL_IN5 / 100000);
    min = (int)(sensor_data.in_min[5] * 16 * NORMAL_IN5 / 100000);
    len += PRINT_VALUES(LM78_IN5_STRING, v, max, min);
#else
    len += sprintf(buf+len, LM78_IN5_STRING ":\tnot supported\n" );
#endif

#ifdef LM78_IN6_STRING
    v = (int)(sensor_data.in[6] * 16 * NORMAL_IN6 / 100000);
    max = (int)(sensor_data.in_max[6] * 16 * NORMAL_IN6 / 100000);
    min = (int)(sensor_data.in_min[6] * 16 * NORMAL_IN6 / 100000);
    len += PRINT_VALUES(LM78_IN6_STRING, v, max, min);
#else
    len += sprintf(buf+len, LM78_IN6_STRING ":\tnot supported\n" );
#endif

    v = sensor_data.interrupt_status[0];
    len += sprintf(buf+len, "Interrupts:");

    if(v & 0x01) len += sprintf(buf+len, " " LM78_IN0_STRING);
    if(v & 0x02) len += sprintf(buf+len, " " LM78_IN1_STRING);
    if(v & 0x04) len += sprintf(buf+len, " " LM78_IN2_STRING);
    if(v & 0x08) len += sprintf(buf+len, " " LM78_IN3_STRING);
    if(v & 0x10) len += sprintf(buf+len, " Temperature");
    if(v & 0x20) len += sprintf(buf+len, " LM75");
    if(v & 0x40) len += sprintf(buf+len, " " LM78_FAN1_STRING);
    if(v & 0x80) len += sprintf(buf+len, " " LM78_FAN2_STRING);

    v = sensor_data.interrupt_status[1];
    if(v & 0x01) len += sprintf(buf+len, " " LM78_IN4_STRING);
    if(v & 0x02) len += sprintf(buf+len, " " LM78_IN5_STRING);
    if(v & 0x04) len += sprintf(buf+len, " " LM78_IN6_STRING);
    if(v & 0x08) len += sprintf(buf+len, " " LM78_FAN3_STRING);
    if(v & 0x10) len += sprintf(buf+len, " Chassis_Intrusion");
    if(v & 0x20) len += sprintf(buf+len, " FIFO_Overflow");
    if(v & 0x40) len += sprintf(buf+len, " SMI_IN");
    if(v & 0x80) len += sprintf(buf+len, " unknown");
        
    len += sprintf(buf+len, "\n");

#ifdef LM78_VID
    v = sensor_data.vid;

    if (((lm_type == LMTYPE_78) || (lm_type == LMTYPE_78_J)) && (v == 0)) 
      len += sprintf(buf+len,"VID: probably not connected (says 3.5V)\n");
    else {
      if ((lm_type == LMTYPE_78) || (lm_type == LMTYPE_78_J))
        v |= 0x10;
   
      if (v <= 0x01)
          len += sprintf(buf+len, "VID:\t 2.0%dV\n", (0x01 - v) * 5);
      else if (v <= 0x04)
          len += sprintf(buf+len, "VID:\t 1.%dV\n", (0x04 - v) * 5 + 85);
      else if (v <= 0x0f)
          len += sprintf(buf+len, "VID:\t reserved (unknown) value\n");
      else if (v <= 0x15)
          len += sprintf(buf+len, "VID:\t 3.%dV\n", (0x15 - v));
      else if (v <= 0x1e)
          len += sprintf(buf+len, "VID:\t 2.%dV\n", (0x1e - v) + 1);
      else 
          len += sprintf(buf+len, "VID:\t No core installed.\n"); 
    } 
#endif

    len += sprintf(buf+len, "Mainboard temperature: %d degrees (max = %d, "
                   "min = %d)\n", sensor_data.temp_mb, sensor_data.temp_mb_max,
                   sensor_data.temp_mb_min);

    fan = sensor_data.fan[0];
    len += sprintf(buf+len, LM78_FAN1_STRING ":\t");
    if (fan==255)
      len += sprintf(buf+len, "0 rpm (not present?)");
    else
      len += sprintf(buf+len, "%d rpm", LM78_RPM(fan,2));

    len += sprintf(buf+len, " (max = %d rpm)\n",
                   LM78_RPM(sensor_data.fan_max[0], 2));

    fan = sensor_data.fan[1];
    len += sprintf(buf+len, LM78_FAN2_STRING ":\t");
    if (fan==255)
      len += sprintf(buf+len, "0 rpm (not present?)");
    else
      len += sprintf(buf+len, "%d rpm", LM78_RPM(fan,2));

    len += sprintf(buf+len, " (max = %d rpm)\n",
                   LM78_RPM(sensor_data.fan_max[1], 2));

    fan = sensor_data.fan[2];
    len += sprintf(buf+len, LM78_FAN3_STRING ":\t");
    if (fan==255)
      len += sprintf(buf+len, "0 rpm (not present?)");
    else
      len += sprintf(buf+len, "%d rpm", LM78_RPM(fan,2));

    len += sprintf(buf+len, " (max = %d rpm)\n",
                   LM78_RPM(sensor_data.fan_max[0], 2));
  }
  return len;
}


/* Read a value from the LM78. Theoretically, several drivers can use
   the LM78 at once by using this function. In practice, this is unlikely,
   but still it won't do any harm. */
u8 lm78_read_value(u8 address) 
{
  while (LM78_BUSY) {
    current->state = TASK_INTERRUPTIBLE;
    current->timeout = jiffies + 10;
    schedule();
  }
  return LM78_READ(address);
}

/* Reread the sensor information into struct sensor_data. */
void update_lm78_values(void)
{
  /* TODO: grab Spin lock on sensor_data */

#ifdef LM78_IN0_STRING
  sensor_data.in[0] = lm78_read_value(LM78_IN0);
  sensor_data.in_max[0] = lm78_read_value(LM78_IN0_MAX);
  sensor_data.in_min[0] = lm78_read_value(LM78_IN0_MIN);
#endif

#ifdef LM78_IN1_STRING
  sensor_data.in[1] = lm78_read_value(LM78_IN1);
  sensor_data.in_max[1] = lm78_read_value(LM78_IN1_MAX);
  sensor_data.in_min[1] = lm78_read_value(LM78_IN1_MIN);
#endif

#ifdef LM78_IN2_STRING
  sensor_data.in[2] = lm78_read_value(LM78_IN2);
  sensor_data.in_max[2] = lm78_read_value(LM78_IN2_MAX);
  sensor_data.in_min[2] = lm78_read_value(LM78_IN2_MIN);
#endif

#ifdef LM78_IN3_STRING
  sensor_data.in[3] = lm78_read_value(LM78_IN3);
  sensor_data.in_max[3] = lm78_read_value(LM78_IN3_MAX);
  sensor_data.in_min[3] = lm78_read_value(LM78_IN3_MIN);
#endif

#ifdef LM78_IN4_STRING
  sensor_data.in[4] = lm78_read_value(LM78_IN4);
  sensor_data.in_max[4] = lm78_read_value(LM78_IN4_MAX);
  sensor_data.in_min[4] = lm78_read_value(LM78_IN4_MIN);
#endif

#ifdef LM78_IN5_STRING
  sensor_data.in[5] = lm78_read_value(LM78_IN5);
  sensor_data.in_max[5] = lm78_read_value(LM78_IN5_MAX);
  sensor_data.in_min[5] = lm78_read_value(LM78_IN5_MIN);
#endif

#ifdef LM78_IN6_STRING
  sensor_data.in[6] = lm78_read_value(LM78_IN6);
  sensor_data.in_max[6] = lm78_read_value(LM78_IN6_MAX);
  sensor_data.in_min[6] = lm78_read_value(LM78_IN6_MIN);
#endif

#ifdef LM78_VID
  if (! sensor_data.valid) {
    sensor_data.vid = lm78_read_value(LM78_VID) & 0x0f;
    if(lm_type == LMTYPE_79)
      sensor_data.vid |= ((lm78_read_value(LM78_RESET) & 0x01) << 4);
  }
#endif

  sensor_data.temp_mb = lm78_read_value(LM78_TEMPERATURE);
  sensor_data.temp_mb_max = lm78_read_value(LM78_TEMPERATURE_MAX);
  sensor_data.temp_mb_min = lm78_read_value(LM78_TEMPERATURE_MIN);

  sensor_data.fan[0] = lm78_read_value(LM78_FAN1);
  sensor_data.fan[1] = lm78_read_value(LM78_FAN2);
  sensor_data.fan[2] = lm78_read_value(LM78_FAN3);

  sensor_data.fan_max[0] = lm78_read_value(LM78_FAN1_MAX);
  sensor_data.fan_max[1] = lm78_read_value(LM78_FAN2_MAX);
  sensor_data.fan_max[2] = lm78_read_value(LM78_FAN3_MAX);

  sensor_data.interrupt_status[0] = lm78_read_value(LM78_INTSTAT1);
  sensor_data.interrupt_status[1] = lm78_read_value(LM78_INTSTAT2);

  /* TODO: release Spin lock on sensor_data */
}
