/* -*- linux-c -*- */


/*
    SMBus.c - Part of a Linux module for reading sensor data.
    Copyright (c) 1998  Philip Edelbrock <phil@netroedge.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/*
 *  SMBus driver for LINUX
 *
 * Copyright (c) 1998 Philip Edelbrock
 * 
 * part of a module which provides a 
 * /proc/sensors file which shows
 * some information about your system.
 * Voltages, temperature and fan speeds
 * are currently reported.
 *
 *
 * please report any bugs/patches to me:
 *  Philip Edelbrock <phil@netroedge.com>
 *
 *
 * Reference:
 *   National Semiconductor
 *   LM75 Digital Temperature Sensor and Thermal Watchdog with
 *   Two-Wire Interface, PDF, April 1997.
 *   http://www.national.com/ds/LM/LM75.pdf
 *
 *   Intel
 *   82371AB PCI-TO-ISA / IDE XCELERATOR (PIIX4), PDF, April 1997.
 *   ftp://ftp.intel.com
 *   
 *   System Management Bus Specification, Rev. 1.0, 1996.
 *   Contributors:
 *   Benchmarq Microelectronics Inc ., Duracell Inc.,
 *   Energizer Power Systems, Intel Corporation, Linear Technology Corporation,
 *   Maxim Integrated Products, Mitsubishi Electric Corporation,
 *   National Semiconductor Corporation, Toshiba Battery Co.,
 *   Varta Batterie AG, All rights reserved.
 *   http://www.sbs-forum.org/
 */
#include <linux/config.h>
#include <linux/version.h>

/* The next define disables the redefinition of kernel_version. As it is
   already defined and initialized in lm78.c, this would result in twice
   defined symbols. We must include <linux/version.h> by hand now. */
#define __NO_VERSION__

/* Hack to allow the use of CONFIG_MODVERSIONS. In the kernel itself, this
   is arranged from the main Makefile. */
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include "smbus.h"
#include "main.h"

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0))
#include <linux/bios32.h>
#endif

/*
	Global variables
*/

unsigned short PIIX4_smba = 0;
int smbus_initialized=0;


/* This should be severly cleaned up, ie. more code which works for both
   2.0 and 2.1 kernels. */
#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,1,0))
/*
 *      Initialize SMBus code.
 */
int Init_SMBus(void){
int error_return=0;

struct pci_dev *PIIX4_dev;

#ifdef DEBUG
unsigned char temp;
#endif

    if (pci_present() == 0) {
        printk("LM78: SMBus: Error: No PCI-bus found!\n");
        error_return=1;
    } else {    

        PIIX4_dev = pci_find_device(PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82371AB_0, NULL);
        if(PIIX4_dev == NULL) {
            printk("LM78: Error: Can't detect PIIX4!\n");
            error_return=1;
        } else {
            /* not sure why this line doesn't work, but it doesn't on my machine
               try it on yours (instead of the next line)
               pci_read_config_word(PIIX4_dev, SMBBA, &PIIX4_smba);*/
            pcibios_read_config_word(PIIX4_dev->bus->number, PIIX4_dev->devfn | 3, SMBBA, &PIIX4_smba);
            PIIX4_smba &= 0xfff0;   
            if (check_region(PIIX4_smba, 8)) {
	    	printk("LM78: SMBus: Error: PIIX4_smb region 0x%x already in use!\n", PIIX4_smba);
                error_return=1; 
	    } else {
	    
	    /* Everything is happy, let's grab the memory and set things up. */
            	request_region(PIIX4_smba, 8, "lm78");
		smbus_initialized=1;
#ifdef DEBUG

                pci_read_config_byte(PIIX4_dev, SMBHSTCFG, &temp);
		if ((temp & 1) == 0)
			printk("LM78: SMBUS: Error: Host SMBus controller not enabled!\n");
		if ((temp & 0x0E) == 8)
			printk("LM78: SMBUS: PIIX4 using Interrupt 9 for SMBus.\n");
		else
			if ((temp & 0x0E) == 0)
				printk("LM78: SMBUS: PIIX4 using Interrupt SMI# for SMBus.\n");
			else
				printk(
			"LM78: SMBUS: PIIX4: Illegal Interrupt configuration (or code out of date)!\n");

		pci_read_config_byte(PIIX4_dev, SMBREV, &temp);
		printk("LM78: SMBUS: SMBREV = 0x%X\n",temp);
#endif /* DEBUG */
	    }
        }   
    }   
    if (error_return)
        printk("LM78: SMBus setup failed.\n");
#ifdef DEBUG
    else
    	printk("PIIX4_smba = 0x%X\n",PIIX4_smba);
#endif
    return error_return;
}
#else
/*
 *      Initialize SMBus code.
 */
int Init_SMBus(void){
int error_return=0;
unsigned char  PIIX4_bus, PIIX4_devfn;
#ifdef DEBUG
unsigned char temp;
#endif

    if (pcibios_present() == 0) {
        printk("LM78: SMBus: Error: No PCI-bus found!\n");
        error_return=1;
    } else {
        if(pcibios_find_device(PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82371AB_0, 0,
                               &PIIX4_bus, &PIIX4_devfn)) {
            printk("LM78: Error: Can't detect PIIX4!\n");
            error_return=1;
        } else {
            pcibios_read_config_word(PIIX4_bus, PIIX4_devfn | 3, SMBBA, &PIIX4_smba);
            PIIX4_smba &= 0xfff0;
            if (check_region(PIIX4_smba, 8)) {
                printk("LM78: SMBus: Error: PIIX4_smb already in use!\n");
                error_return=1;
            } else {

            /* Everything is happy, let's grab the memory and set things up. */
                request_region(PIIX4_smba, 8, "lm78");
                smbus_initialized=1;
#ifdef DEBUG
                pcibios_read_config_byte(PIIX4_bus, PIIX4_devfn | 3, SMBHSTCFG,
&temp);
                if ((temp & 1) == 0)
                        printk("LM78: SMBUS: Error: Host SMBus controller not enabled!\n");
                if ((temp & 0x0E) == 8)
                        printk("LM78: SMBUS: PIIX4 using Interrupt 9 for SMBus.\n");
                else
                        if ((temp & 0x0E) == 0)
                                printk("LM78: SMBUS: PIIX4 using Interrupt SMI#
for SMBus.\n");
                        else
                                printk(
                        "LM78: SMBUS: PIIX4: Illegal Interrupt configuration (or code out of date)!\n");
                pcibios_read_config_byte(PIIX4_bus, PIIX4_devfn | 3, SMBREV, &temp);
                printk("LM78: SMBUS: SMBREV = 0x%X\n",temp);
#endif
            }
        }
    }
    if (error_return)
        printk("LM78: SMBus setup failed.\n");
#ifdef DEBUG
    else
        printk("PIIX4_smba = 0x%X\n",PIIX4_smba);
#endif
    return error_return;
}
#endif

/*
 *      Clean-up the SMBus code.
 */
void Cleanup_SMBus(void){
        release_region(PIIX4_smba, 8); 
	smbus_initialized=0;
}

/*
 *      Software pause (busy-wait).
 */
static void do_pause( unsigned amount )
{
   mdelay(amount * 10);
}

/*
 *      Write byte value to SMBus device at given address.
 */
int write_smbus(int addr, long value, int size) {

int HOST_BUSY;
int timeout=0;
int result=0;

    if (! smbus_initialized) {
       printk("SMBus: Attempt made to use uninitalized SMBus routines!\n");
       return -1;
    }
/* Setup protocol type, address, device command, data, and control bits */
    outb_p( value & 0x0FF, SMBHSTCMD);
    outb_p( (value & 0x0FF00) >> 8, SMBHSTDAT0);
    outb_p( (value & 0x0FF0000) >> 16, SMBHSTDAT1); /* Not nessesary, but doesn't hurt */
    outb_p( (size & 0x1C) + (ENABLE_INT9 & 1), SMBHSTCNT); /* SMBus operation size and Int config */
    if (inb_p(SMBHSTSTS) != 0x00) {
#ifdef DEBUG
	printk("LM78: write_smbus_byte: SMBus busy. Resetting... ");
#endif
        outb_p( inb(SMBHSTSTS), SMBHSTSTS);
#ifdef DEBUG
	if (inb_p(SMBHSTSTS) != 0x00)
		printk("Failed!\n");
	else
		printk("Successfull!\n");
#endif
    }
    outb_p( (addr << 1) & 0x0FE, SMBHSTADD);
/* address of possible thermal sensors are from
 1001000x to 1001111x ( x = r/w bit ) */

/* Set the start bit */
    outb_p(inb(SMBHSTCNT) | 0x040, SMBHSTCNT); 
/* start the transaction by setting bit 6 */

/* Wait for a fraction of a second! */
   do_pause(10);

/* Poll Host_busy bit */
   HOST_BUSY=inb_p(SMBHSTSTS) & 0x01;
   while (HOST_BUSY & (timeout++ < MAX_TIMEOUT)) {
/* Wait for a while and try again*/
        do_pause(5);
	HOST_BUSY=inb_p(SMBHSTSTS) & 0x01;
   }
#ifdef DEBUG
   if (timeout++ >= MAX_TIMEOUT)
	   printk("LM78: write_smbus_byte: SMBus Timeout!\n");
   printk("LM78: write_smbus_byte: time cycles for transaction: %d (0x%X)\n",timeout,addr);
#endif

/* Sense for no device at the address */
   if (inb_p(SMBHSTSTS) == 0x04)  {
        result=-1;
#ifdef DEBUG
	printk("LM78: write_smbus_byte: SMBus error: no response from 0x%X !\n",addr);
#endif
   }
#ifdef DEBUG
   if (inb_p(SMBHSTSTS) & 0x010)
	printk("LM78: write_smbus_byte: SMBus error: Failed bus transaction (0x%X).\n",addr);
   if (inb_p(SMBHSTSTS) & 0x08)
	printk("LM78: write_smbus_byte: SMBus error: Bus collision (0x%X).\n",addr);
   if (inb_p(SMBHSTSTS) & 0x04)
	printk("LM78: write_smbus_byte: SMBus error: Device Error (0x%X).\n",addr);
#endif
    if (inb_p(SMBHSTSTS) != 0x00)
        outb_p( inb(SMBHSTSTS), SMBHSTSTS);
   return result;
}


/*
 *      Read value from SMBus device at given address.
 */
int read_smbus(int addr, int command, int size) {

int HOST_BUSY;
int timeout=0;
int result=0;

    if (! smbus_initialized) {
       printk("SMBus: Attempt made to use uninitalized SMBus routines!\n");
       return -1;
    }
/* Setup protocol type, address, device command, data, and control bits */
    outb_p( command & 0x0FF, SMBHSTCMD);
    outb_p( 0x00, SMBHSTDAT0);
    outb_p( 0x00, SMBHSTDAT1);
    outb_p( (size & 0x1C) + (ENABLE_INT9 & 0x01), SMBHSTCNT); /* SMBus operation size and interrupt.
    							Everything else should be zero for now */
    if (inb_p(SMBHSTSTS) != 0x00) {
        outb_p( inb(SMBHSTSTS), SMBHSTSTS);
    }
    outb_p( (addr << 1) | 0x01, SMBHSTADD);
/* address of possible thermal sensors are from
 1001000x to 1001111x ( x = r/w bit ) */

/* Set the start bit */
    outb_p(inb(SMBHSTCNT) | 0x040, SMBHSTCNT); 
/* start the transaction by setting bit 6 */

/* Wait for a fraction of a second! */
   do_pause(10);

/* Poll Host_busy bit */
   HOST_BUSY=inb_p(SMBHSTSTS) & 0x01;
   while (HOST_BUSY & (timeout++ < MAX_TIMEOUT)) {
/* Wait for a while and try again*/
        do_pause(5);
	HOST_BUSY=inb_p(SMBHSTSTS) & 0x01;
   }
   
#ifdef DEBUG
   if (timeout++ >= MAX_TIMEOUT)
	   printk("LM78: read_smbus: SMBus Timeout!\n");
   printk("LM78: read_smbus: time cycles for transaction: %d (0x%X)\n",timeout,addr);
#endif

   if (! (inb(SMBHSTSTS) & 0x01)) {
        result=(inb(SMBHSTDAT0) << 8) | inb(SMBHSTDAT1);
   }

/* Sense for no device at the address */
   if (inb_p(SMBHSTSTS) == 0x04)  {
        result=-1;
#ifdef DEBUG
	printk("LM78: read_smbus: SMBus error: no response from 0x%X !\n",addr);
#endif
   }
#ifdef DEBUG
   if (inb_p(SMBHSTSTS) & 0x010)
	printk("LM78: read_smbus: SMBus error: Failed bus transaction (0x%X).\n",addr);
   if (inb_p(SMBHSTSTS) & 0x08)
	printk("LM78: read_smbus: SMBus error: Bus collision (0x%X).\n",addr);
   if (inb_p(SMBHSTSTS) & 0x04)
	printk("LM78: read_smbus: SMBus error: Device Error (0x%X).\n",addr);
#endif
   if (inb_p(SMBHSTSTS) != 0x00)
        outb_p( inb(SMBHSTSTS), SMBHSTSTS);
   return result;
}
